/*
 * Created on Nov 3, 2004
 *
 */
package com.sunlabs.archsim.engine;

import com.sunlabs.archsim.base.simulation.Event;




/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * @author ib27688
 *
 * EventHeap implements a heap data structure to be used in the event priority queue
 */
/*
 * I decided not to bother with a generic heap. 
 */
public class EventHeap {

	/** Array to hold events, organized as a heap */
	protected Event event[];
	
	/** 	Default arrray size whenusing a default constructor */
	protected final static int DefaultCapacity= 100;
	
	/** The number of items in the heap */
	protected int nextEmptyPlace= 0;
	
	/** create a default-size array to hold the heap */
	public EventHeap() {
		event= new Event[DefaultCapacity]; 
	} // EventHeap default constructor
	
	/** Create a specified size of array to hold the heap. Do not go below default size */
	public EventHeap(int initialSize) {
		if (initialSize < DefaultCapacity) {
			event= new Event[DefaultCapacity];
		}
		else {
			event= new Event[initialSize];
		}
	} // EventHeap constructor
	
	/** 
	 * @return return the number of the events in the queue
	 */
	public int size() {
		return nextEmptyPlace;
	} // size
	
	/**
	 * @return the max number of events that  the current queue can hold (queue aray size)
	 */
	public int capacity() {
		return event.length;
	} // capacity
	
	/** 
	 * Resize the heap (array) to the new capacity. 
	 * Do not resize if the new size is smaller than the number of items in the heap.
	 * Do not shrink under the default heap size.
	 * 
	 * @param newCapacity the new size
	 * @return the size of the array after the grow methos
	 */
	public int resize(int newCapacity) {
		// do nothing if the new size is too small
		if (newCapacity >= DefaultCapacity && newCapacity > nextEmptyPlace) {
			Event newArray[]= new Event[newCapacity];
			// copy the events over to the new array
			System.arraycopy(event, 0, newArray, 0, (nextEmptyPlace<event.length)?nextEmptyPlace : event.length);
			// switch over to the new array
			event= newArray;
		}
		return event.length;
		
	} // resize
	
	/**
	 * Assume the standard array representation of a heap, element 0 contains the root
	 * @param node index of a node in a heap
	 * @return index of node's parent
	 */
	protected int parent(int node) {
		return (node>0)? (node-1)/2 : 0;
	} // parent
	
	/**
	 * Assume the standard array representation of a heap, element 0 contains the root
	 * @param node node index in a heap
	 * @return index of the left child of the node
	 */
	protected int leftChild(int node) {
		return node*2 +1;
	} // left child
	
	/**
	 * Assume the standard array representation of a heap, element 0 contains the root
	 * @param node index of a node in a heap
	 * @return index of the right child
	 */
	protected int rightChild(int node) {
		return node*2+2;
	} // rightChild
	
	
	/**
	 * Return and remove the top event from the queue.
	 * @returns the top event from the heap
	 */
	public Event getTop() {
		if (nextEmptyPlace == 0) return null;
		// record the top
		Event top= event[0];
		event[0]= event[nextEmptyPlace-1]; // move the last element to the top
		event[nextEmptyPlace-1]= null;     // null-ify the last item, not really needed, cleaner for debugging
		nextEmptyPlace= nextEmptyPlace-1;  // the heap has decreased
		
		// now we have to re-establish the heap - the left and right subtree of the root are both
		// heaps. we need to replace the root with the smaller of the two children and repet the
		// process in the corresponding subtree until the whole heap is reestablished
		int current= 0;  // current position of the "sinking" root
		int earliest= 0; // the position of the earliest of the "sinking root", left and right child;
		while (true) {
			int lC= leftChild(current); // find the positions of the children of the root of the current tree
			int rC= rightChild(current);
			
			// should the left child be pulled up?
			if (lC < nextEmptyPlace && 
					event[earliest].getTime().laterThan(event[lC].getTime())) {
				earliest= lC;
			}
			// should the right child be pulled up?
			if (rC < nextEmptyPlace && 
					event[earliest].getTime().laterThan(event[rC].getTime())) {
				earliest= rC;
			}
			if (current != earliest) {
				Event tmp= event[current];
				event[current]= event[earliest];
				event[earliest]= tmp;
				current= earliest;
			}
			else break; // the "sinking root" found its place
		} // while
		// now return what was on the top of the heap when we started
		return top;
	} // getTop
	
	/**
	 * Take a peek at  the top event on the queue, do not affect the queue
	 * @return top event from the queuer
	 */
	public Event peek() {
		return event[0];
	} // peek
	
	/**
	 * Insert a new event in the heap
	 * @param newEvent the new event to be inserted
	 */
	public void insert(Event newEvent) {
		if (newEvent != null) {  
			// do we have space? if not, resize
			if (nextEmptyPlace >= event.length) resize(event.length + DefaultCapacity);
			
			// mark that a new event wil be inserted
			nextEmptyPlace++;
			
			// propagate the new event up the tree until it's place is found
			int target= nextEmptyPlace-1;
			// move the new event up the tree while the event comes before its parent
			while (target > 0 && newEvent.getTime().beforeThan(event[parent(target)].getTime())) {
				event[target]= event[parent(target)];
				target= parent(target);
			}
			// found a place for new event
			event[target]= newEvent;			
		}
	} // insert
	
	/**
	 * clear the contents of the heap
	 */
	public void clear() {
		for (int i= 0; i< nextEmptyPlace; i++) event[i]= null;
		nextEmptyPlace= 0;
	} // clear
	
	public void print() {
		for (int i=0; i< nextEmptyPlace; i++) {
			System.out.println(event[i]);
		}
	} // print
	
	
} // class EventHeap
